"""
yubistack.wsgi
~~~~~~~~~~~~~~

WSGI wrappers around the yubistack functions
to support backward compatibility.
"""

import json
import logging
import re
import time

from yubistack.config import (
    settings,
    TOKEN_LEN
)
from yubistack.exceptions import (
    YKAuthError,
    YKValError,
    YKSyncError,
    YKKSMError,
)
from yubistack.utils import (
    parse_querystring,
    wsgi_response,
    sign,
)
from yubistack.ykauth import Client
from yubistack.ykksm import Decryptor
from yubistack.yksync import Sync
from yubistack.ykval import Validator

if settings['SYSLOG_WSGI_AUTH']:
    import syslog
    syslog.openlog(ident='yubistack', facility=syslog.LOG_AUTH)

logger = logging.getLogger(__name__)

HTTP_STATUS_CODES = {
    200: '200 OK',
    400: '400 Bad Request',
    401: '401 Unauthorized',
    404: '404 Not Found',
    500: '500 Internal Server Error',
}
URI_REGEX = re.compile(r"""
    ^/?                     # Leading slash might not be present
    (                       # Intersection: yubiauth and yubikey-val + yubikey-ksm differences
        wsapi(/\d+\.\d+)?   # YKVal and YKKsm uses wsapi prefix + protocol version
      | yubiauth/client     # Yubiclient authentication URI
    )
    /
    (?P<resource>           # The resource should map to a function name
        (
            decrypt
          | verify
          | sync
          | resync
          | authenticate
        )
    )
    /?                      # Trailing clash might not be present
""", re.VERBOSE)

PERSISTENT_OBJECTS = {}

REQUIRED_AUTH_PARAMS = ['username', 'password', 'otp']
def authenticate(environ, start_response):
    """
    Handle authentications
    """
    start_time = time.time()

    if 'client' not in PERSISTENT_OBJECTS:
        PERSISTENT_OBJECTS['client'] = Client()
    client = PERSISTENT_OBJECTS['client']
    _format = 'json' if environ.get('HTTP_ACCEPT') == 'application/json' else 'text'
    params = {}

    try:
        # Parse POST request
        try:
            request_body_size = int(environ.get('CONTENT_LENGTH', 0))
        except ValueError:
            request_body_size = 0
        request_body = environ['wsgi.input'].read(request_body_size)
        params = parse_querystring(request_body.decode())
        _params = params.copy()
        _params['password'] = '*' * 8
        logger.debug('PROCESSED QUERYSTRING: %s', _params)
        # Checking parameters
        for req_param in REQUIRED_AUTH_PARAMS:
            if req_param not in params:
                raise YKAuthError('MISSING_PARAMETER')
        client.authenticate(params['username'], params['password'], params['otp'])
        status_code = 200
        output = {'status': 'OK',
                  'message': 'Successful authentication'}
    except (YKAuthError, YKValError, YKSyncError, YKKSMError) as err:
        status_code = 400
        output = {'status': err.error_code,
                  'message': str(err)}
    except Exception as err:
        status_code = 500
        output = {'status': 'BACKEND_ERROR',
                  'message': 'Backend error: %s' % err}
        logger.exception('Backend error: %s', err)
    finally:
        content_type = 'application/json' if _format == 'json' else 'text/plain'
        start_response(HTTP_STATUS_CODES[status_code], [('Content-Type', content_type)])
        output['username'] = params.get('username', '')
        output['token_id'] = params.get('otp', '')[:-TOKEN_LEN]
        output['latency'] = round(time.time() - start_time, 3)
        output['src_ip'] = environ.get('REMOTE_ADDR')
        response = json.dumps(output if _format == 'json' else (status_code == 200))
        if status_code == 200:
            logger_sev = logging.INFO
            syslog_sev = syslog.LOG_INFO
        elif status_code == 400:
            logger_sev = logging.WARNING
            syslog_sev = syslog.LOG_WARNING
        else:
            logger_sev = logging.ERROR
            syslog_sev = syslog.LOG_ERR
        if settings['SYSLOG_WSGI_AUTH']:
            syslog.syslog(syslog_sev, response if _format == 'json' else json.dumps(output))
        logger.log(logger_sev, '[%(username)s][%(token_id)s] %(status)s: %(message)s', output)
        return [response.encode()]

def decrypt(environ, start_response):
    """
    Handle OTP decryptions
    """
    _format = 'text'
    try:
        params = parse_querystring(environ['QUERY_STRING'])
        if params.get('format') == 'json' or environ['HTTP_ACCEPT'] == 'application/json':
            _format = 'json'
        logger.debug('PROCESSED QUERYSTRING: %s', params)
        decryptor = Decryptor()
        output = decryptor.decrypt(params.get('otp'))
        if _format != 'json':
            output = 'OK counter=%(counter)s low=%(low)s high=%(high)s use=%(use)s\n' % output
        status_code = 200
    except YKKSMError as err:
        logger.exception('Decryption error: %s', err.error_code)
        output = '%s' % str(err)
        status_code = 400
    except Exception as err:
        logger.exception('Backend error: %s', err)
        output = 'Backend failure\n'
        status_code = 500
    finally:
        content_type = 'application/json' if _format == 'json' else 'text/plain'
        if _format == 'json':
            if isinstance(output, str):
                output = {'error': output}
            output = json.dumps(output)
        elif status_code != 200:
            output = 'ERR %s\n' % output
        start_response(HTTP_STATUS_CODES[status_code], [('Content-Type', content_type)])
        return [output.encode()]

PARAM_MAP = {
    'id': 'client_id',
    'sl': 'sync_level',
}
def verify(environ, start_response):
    """
    Handle OTP Validation
    """
    apikey = ''.encode()
    try:
        params = parse_querystring(environ['QUERY_STRING'])
        public_id = params.get('otp', '?' * 12)[:12]
        logger.debug('%s: PROCESSED QUERYSTRING: %s', public_id, params)
        validator = Validator()
        apikey = validator.get_client_apikey(params.get('id'))
        client_signature = params.pop('h')
        server_signature = sign(params, apikey)
        if client_signature != server_signature:
            logger.error('[%s] Client hmac=%s != Server hmac=%s',
                         public_id, client_signature, server_signature)
            raise YKValError('BAD_SIGNATURE')
        for old_key, new_key in PARAM_MAP.items():
            if old_key in params:
                params[new_key] = params[old_key]
                params.pop(old_key)
        extra = validator.verify(**params)
        output = 'OK'
        logger.info('[%s] OTP Verified', public_id)
    except YKValError as err:
        output = '%s' % err
    except Exception as err:
        logger.exception('%s: Backend error: %s', public_id, err)
        output = 'BACKEND_ERROR'
    finally:
        return wsgi_response(output, start_response, apikey=apikey, extra=None)

def sync(environ, start_response):
    """
    Handle Sync requests
    """
    local_params = None
    try:
        # Validate caller address
        if environ['REMOTE_ADDR'] not in settings['SYNC_POOL']:
            logger.error('Operation not permitted from IP %(REMOTE_ADDR)s', environ)
            raise YKSyncError('OPERATION_NOT_ALLOWED',
                              'Remote IP %(REMOTE_ADDR)s it not in sync pool' % environ)
        sync_params = parse_querystring(environ['QUERY_STRING'])
        logger.info('[%s] Received sync request from %s (counter: %s, use: %s, nonce: %s)',
                    sync_params.get('yk_publicname'), environ['REMOTE_ADDR'],
                    sync_params.get('yk_counter'), sync_params.get('yk_use'),
                    sync_params.get('nonce'))
        synclib = Sync()
        local_params = synclib.sync_local(sync_params)
        output = 'OK'
        status_code = 200
    except YKSyncError as err:
        output = str(err)
        status_code = 401
    except Exception as err:
        logger.exception('ERROR: %s', err)
        output = 'BACKEND_ERROR'
        status_code = 500
    finally:
        return wsgi_response(output, start_response, apikey=''.encode(),
                             extra=local_params, status=status_code)

def resync(environ, start_response):
    """
    Handle Re-sync requests
    """
    try:
        # Validate caller address
        if environ['REMOTE_ADDR'] not in settings['SYNC_POOL']:
            logger.error('Operation not permitted from IP %(REMOTE_ADDR)s', environ)
            raise YKSyncError('OPERATION_NOT_ALLOWED',
                              'Remote IP %(REMOTE_ADDR)s it not in sync pool' % environ)
        # Parse query and check values
        resync_params = parse_querystring(environ['QUERY_STRING'])
        synclib = Sync()
        output = synclib.resync_local(resync_params)
        status_code = 200
        logger.info('Re-sync request by %s for keys: %s',
                    environ['REMOTE_ADDR'], resync_params['yk'])
    except YKSyncError as err:
        output = str(err)
        status_code = 401
    except Exception as err:
        logger.exception('ERROR: %s', err)
        output = ''
        status_code = 500
    finally:
        start_response(HTTP_STATUS_CODES[status_code], [('Content-Type', 'text/plain')])
        return [output.encode()]

def router(environ, start_response):
    """ Simple WSGI router """
    path = environ.get('PATH_INFO', '')
    match = re.match(URI_REGEX, path)
    if not match or not match.groupdict()['resource']:
        start_response(HTTP_STATUS_CODES[404], [('Content-Type', 'text/plain')])
        return ['Invalid URI'.encode()]
    func = globals()[match.groupdict()['resource']]
    return func(environ, start_response)

def main():
    """ Run a web server to test the application """
    from wsgiref.simple_server import make_server
    logging.basicConfig(format='%(asctime)s %(levelname)s %(module)s %(funcName)s: %(message)s',
                        level=logging.DEBUG)
    srv = make_server('0.0.0.0', 8080, router)
    srv.serve_forever()

if __name__ == '__main__':
    main()



"""
yubistack.ykval
~~~~~~~~~~~~~~~

Python Yubikey Stack - Key validation module
"""

import logging
try:
    import queue
except ImportError:
    import Queue as queue
import re
import threading
import time

import requests

from .config import settings
from .db import DBHandler
from .exceptions import YKSyncError
from .utils import (
    generate_nonce,
    counters_eq,
    counters_gt,
    counters_gte,
    parse_sync_response,
)

logger = logging.getLogger(__name__)
requests_log = logging.getLogger('requests')
requests_log.setLevel(logging.WARNING if logger.getEffectiveLevel() != 10 else logging.DEBUG)

REQUIRED_PARAMS = ['modified', 'otp', 'nonce', 'yk_publicname',
                   'yk_counter', 'yk_use', 'yk_high', 'yk_low']
class Sync(object):
    """ Sync object to handle cross synchronization requests """
    def __init__(self, db=None):
        self.db = db if db else DBHandler(db='ykval')
        self.sync_servers = settings['SYNC_SERVERS']

    def check_sync_input(self, sync_params):
        """ Check for all required parameters """
        for req_param in REQUIRED_PARAMS:
            if req_param not in sync_params:
                logger.error("Received request with missing '%s' parameter", req_param)
                raise YKSyncError('MISSING_PARAMETER', req_param)
            if req_param not in ('otp', 'nonce', 'yk_publicname') and not \
                (sync_params[req_param] == '-1' or isinstance(sync_params[req_param], int)):
                logger.error("Input parameter '%s' is not correct", req_param)
                raise YKSyncError('INVALID_PARAMETER', req_param)

    def check_resync_input(self, resync_params):
        """ Check input parameters """
        if 'yk' not in resync_params:
            logger.error("Received request with missing 'yk' parameter")
            raise YKSyncError('MISSING_PARAMETER', 'yk')
        if not re.match(r'^([cbdefghijklnrtuv]{0,16}|all)$', resync_params['yk']):
            logger.error("Invalid 'yk' value: %(yk)s", resync_params)
            raise YKSyncError('INVALID_PARAMETER', 'yk')

    def sync_local(self, sync_params):
        """ Synchronize """
        self.check_sync_input(sync_params)
        local_params = self.db.get_local_params(sync_params['yk_publicname'])
        self.db.update_db_counters(sync_params)
        logger.debug('[%s] Local params: %s',
                     sync_params['yk_publicname'], local_params)
        logger.debug('[%s] Sync request params: %s',
                     sync_params['yk_publicname'], sync_params)

        if counters_gte(local_params, sync_params):
            logger.warning('[%(yk_publicname)s] Remote server out of sync', sync_params)

        if counters_eq(local_params, sync_params):
            if sync_params['modified'] == local_params['modified'] \
                and sync_params['nonce'] == local_params['nonce']:
                # This is not an error. When the remote server received
                # an OTP to verify, it would have sent out sync requests
                # immediately. When the required number of responses had
                # been received, the current implementation discards all
                # additional responses (to return the result to the client
                # as soon as possible). If our response sent last time was
                # discarded, we will end up here when the background
                # ykval-queue processes the sync request again.
                logger.info('[%(yk_publicname)s] Sync request unnecessarily sent',
                            sync_params)

            if (
                    sync_params['modified'] != local_params['modified'] and
                    sync_params['nonce'] == local_params['nonce']
                ):
                delta_modified = sync_params['modified'] - local_params['modified']
                if delta_modified < -1 or delta_modified > 1:
                    logger.warning('[%s] We might have a replay attack. 2 events '
                                   'at different times have generated the same '
                                   'counters. Time difference is %s sec',
                                   sync_params['yk_publicname'], delta_modified)

            if sync_params['nonce'] != local_params['nonce']:
                logger.warning('[%(yk_publicname)s] Remote server has received '
                               'a request to validate an already validated OTP', sync_params)

        if not local_params['active']:
            # The remote server has accepted an OTP from a YubiKey which
            # we would not. We still needed to update our counters with
            # the counters from the OTP thought.
            logger.error('[%(yk_publicname)s] Received sync-request for '
                         'de-activated Yubikey', sync_params)
            raise YKSyncError('DISABLED_TOKEN')
        return local_params

    def resync_local(self, resync_params):
        """ Re-synchronize """
        self.check_resync_input(resync_params)
        keys = self.db.get_keys(resync_params['yk'])
        server_nonce = generate_nonce()
        for key in keys:
            local_params = self.db.get_local_params(key['yk_publicname'])
            local_params['otp'] = 'c' * 32 # Fake an OTP
            logger.debug('Auth data: %s', local_params)
            for server in self.sync_servers:
                self.db.enqueue(local_params, local_params, server, server_nonce)
        return 'OK Initiated resync of %(yk)s' % resync_params

    def _fetch_remote(self, dqueue, server, url, timeout):
        """ Make HTTP GET call to remote server """
        try:
            req = requests.get(url, timeout=timeout)
            if req.status_code == 200:
                try:
                    resp_params = parse_sync_response(req.text)
                    dqueue.put({'server': server, 'params': resp_params})
                except ValueError as err:
                    logger.error('Failed to parse response of %s: %s', server, err)
            else:
                logger.warning('Recieved status code %s for %s', req.status_code, url)
        except Exception as err:
            logger.warning('Failed to retrieve %s: %s', url, err)

    def sync_remote(self, otp_params, local_params, server_nonce, required_answers, timeout=1):
        """ Function to synchronize values with other ykval servers """
        # Construct URLs
        responses = []
        dqueue = queue.Queue()
        for row in self.db.get_queue(otp_params['modified'], server_nonce):
            url = '%(server)s?otp=%(otp)s&modified=%(modified)s' % row
            url += '&' + row['info'].split(',')[0]
            _thread = threading.Thread(target=self._fetch_remote,
                                       args=(dqueue, row['server'], url, timeout))
            _thread.daemon = True
            _thread.start()
        loop_start = time.time()
        while len(responses) < required_answers and time.time() < loop_start + timeout * 1.5:
            try:
                resp = dqueue.get(timeout=0.2)
                responses.append(resp)
                # Delete entry from table
                self.db.remove_from_queue(resp['server'], otp_params['modified'], server_nonce)
            except queue.Empty:
                pass

        answers = len(responses)
        # Parse response
        valid_answers = 0
        for resp in responses:
            resp_params = resp['params']
            logger.debug('[%(yk_publicname)s] local DB contains %s', otp_params)
            logger.debug('[%s] response contains %s',
                         otp_params['yk_publicname'], resp_params)
            logger.debug('[%s] OTP contains %s',
                         otp_params['yk_publicname'], otp_params)
            # Update Internal DB (conditional)
            self.db.update_db_counters(resp_params)
            # Check for Warnings
            # https://developers.yubico.com/yubikey-val/doc/ServerReplicationProtocol.html
            # NOTE: We use local_params for validationParams comparison since they are actually
            #       the same in this situation and we have them at hand.
            if counters_gt(local_params, resp_params):
                logger.warning('[%(yk_publicname)s] Remote server out of sync', otp_params)
            if counters_gt(resp_params, local_params):
                logger.warning('[%(yk_publicname)s] Local server out of sync', otp_params)
            if counters_eq(resp_params, local_params) \
                and resp_params['nonce'] != local_params['nonce']:
                logger.warning('[%(yk_publicname)s] Servers out of sync. '
                               'Nonce differs.', otp_params)
            if counters_eq(resp_params, local_params) \
                and resp_params['modified'] != local_params['modified']:
                logger.warning('[%(yk_publicname)s] Servers out of sync. '
                               'Modified differs.', otp_params)
            if counters_gt(resp_params, otp_params):
                logger.warning('[%(yk_publicname)s] OTP is replayed. '
                               'Sync response counters higher than OTP counters.', otp_params)
            elif counters_eq(resp_params, otp_params) \
                and resp_params['nonce'] != otp_params['nonce']:
                logger.warning('[%(yk_publicname)s] OTP is replayed. Sync '
                               'response counters equal to OTP counters and nonce '
                               'differs.', otp_params)
            else:
                # The answer is ok since a REPLAY was not indicated
                valid_answers += 1
                if required_answers == valid_answers:
                    break

        # NULL queued_time for remaining entries in queue, to allow
        # daemon to take care of them as soon as possible.
        self.db.null_queue(server_nonce)
        return {'answers': answers, 'valid_answers': valid_answers}



"""
ykval.db
~~~~~~~~

Database Handler & queries
"""

import logging
import time

from .config import settings

logger = logging.getLogger(__name__)

class DBHandler:
    """ Database handler wrapper """
    def __init__(self, db):
        self.settings = settings['DATABASES'][db]
        # This is not nice, but works well.
        # We need to catch OperationalError in the _execute functon, so
        # the db driver has to be available
        if self.settings.get('ENGINE', 'mysql') == 'mysql':
            import MySQLdb as dbdriver
        elif self.settings['ENGINE'] == 'postgres':
            import psycopg2 as dbdriver
        elif self.settings['ENGINE'] == 'sqlite':
            import sqlite3 as dbdriver
        else:
            raise ValueError('Invalid Database configuration')
        self.dbdriver = dbdriver
        self._connect()

    def _connect(self):
        """ Connect to the database """
        if self.settings.get('ENGINE', 'mysql') == 'mysql':
            self._db = self.dbdriver.connect(self.settings['HOST'],
                                             self.settings['USER'],
                                             self.settings['PASSWORD'],
                                             self.settings['NAME'],
                                             self.settings.get('PORT', 3306))
        elif self.settings['ENGINE'] == 'postgres':
            self._db = self.dbdriver.connect(database=self.settings['NAME'],
                                             user=self.settings['USER'],
                                             password=self.settings['PASSWORD'],
                                             host=self.settings['HOST'])
        elif self.settings['ENGINE'] == 'sqlite':
            self._db = self.dbdriver.connect(self.settings['NAME'])
        self.cursor = self._db.cursor()

    def _execute(self, query, params=None, retry=False):
        """ Abstract the cursor execute function to handle sqlite syntax """
        if self.settings.get('ENGINE') == 'sqlite':
            if '%s' in query and params:
                query = query.replace('%s', '?')
            elif '%(' in query and params:
                query = query.replace('%(', ':').replace(')s', '')
        if logger.getEffectiveLevel() == logging.DEBUG:
            _query = ' '.join([x.strip() for x in query.split()])
            logger.debug('QUERY: %s PARAMS: %s', _query, params)
        try:
            rowcount = self.cursor.execute(query, params)
            self._db.commit()
            return rowcount
        except (AttributeError, self.dbdriver.OperationalError) as err:
            if not retry:
                logger.debug('Database reconnect due to error: %s', err)
                self._connect()
                return self._execute(query, params, retry=True)
            else:
                raise
        except Exception as err:
            logger.exception('Database error: %s', err)
            raise

    def _dictfetchall(self):
        """ Wrapper to return DB results in dict format """
        return [dict(zip([col[0] for col in self.cursor.description], row)) \
                for row in self.cursor.fetchall()]

    def _dictfetchone(self):
        """ Wrapper to return DB results in dict format """
        data = self._dictfetchall()
        if data:
            return data[0]
        return {}

    #################
    # YKAUTH QUERIES
    #################
    def get_user(self, username):
        """
        Read user information for Yubiauth
        """
        query = """SELECT users.attribute_association_id AS users_attribute_association_id,
                          users.id AS users_id, users.name AS users_name,
                          users.auth AS users_auth
                     FROM users
                    WHERE users.name = %s"""
        self._execute(query, (username,))
        return self._dictfetchone()

    def get_token(self, user_id, token_id):
        """
        Read user attribute information for Yubiauth
        """
        query = """SELECT yubikeys.attribute_association_id AS yubikeys_attribute_association_id,
                          yubikeys.id AS yubikeys_id,
                          yubikeys.prefix AS yubikeys_prefix,
                          yubikeys.enabled AS yubikeys_enabled
                     FROM yubikeys
               INNER JOIN user_yubikeys
                       ON user_yubikeys.yubikey_id = yubikeys.id
                    WHERE user_yubikeys.user_id = %s
                      AND yubikeys.prefix = %s"""
        self._execute(query, (user_id, token_id))
        return self._dictfetchone()

    #########################
    # YKVAL / YKSYNC QUERIES
    #########################
    def get_client_data(self, client_id):
        """ Lookup client based on the ID """
        query = """SELECT id,
                          secret
                     FROM clients
                    WHERE active = 1
                      AND id = %s"""
        self._execute(query, (client_id,))
        return self._dictfetchone()

    def get_local_params(self, yk_publicname):
        """ Get yubikey parameters from DB """
        query = """SELECT active,
                          modified,
                          yk_publicname,
                          yk_counter,
                          yk_use,
                          yk_low,
                          yk_high,
                          nonce
                     FROM yubikeys
                    WHERE yk_publicname = %s"""
        self._execute(query, (yk_publicname,))
        local_params = self._dictfetchone()
        if not local_params:
            local_params = {
                'active': '1',
                'modified': -1,
                'yk_publicname': yk_publicname,
                'yk_counter': -1,
                'yk_use': -1,
                'yk_low': -1,
                'yk_high': -1,
                'nonce': '0000000000000000',
                'created': int(time.time())
            }
            # Key was missing in DB, adding it
            self.add_new_identity(local_params)
            logger.warning('[%s] Discovered new identity, creating yubikey', yk_publicname)
        logger.debug('[%s] Auth data: %s', yk_publicname, local_params)
        return local_params

    def add_new_identity(self, identity):
        """ Create new key identity """
        query = """INSERT INTO yubikeys (
                       active,
                       created,
                       modified,
                       yk_publicname,
                       yk_counter,
                       yk_use,
                       yk_low,
                       yk_high,
                       nonce
                ) VALUES (
                       %(active)s,
                       %(created)s,
                       %(modified)s,
                       %(yk_publicname)s,
                       %(yk_counter)s,
                       %(yk_use)s,
                       %(yk_low)s,
                       %(yk_high)s,
                       %(nonce)s
                )"""
        self._execute(query, identity)

    def get_queue(self, modified, server_nonce):
        """
        Read all elements from queue
        """
        query = """SELECT server,
                          otp,
                          modified,
                          info
                     FROM queue
                    WHERE modified=%s
                      AND server_nonce = %s"""
        self._execute(query, (modified, server_nonce))
        return self._dictfetchall()

    def read_queue(self):
        """
        Read all elements from queue
        """
        query = """SELECT server,
                          otp,
                          modified,
                          info,
                          server_nonce
                     FROM queue"""
        self._execute(query)
        return self._dictfetchall()

    def remove_from_queue(self, server, modified, server_nonce):
        """
        Remove a single element from queue
        """
        query = """DELETE FROM queue
                         WHERE server = %s
                           AND modified = %s
                           AND server_nonce = %s"""
        self._execute(query, (server, modified, server_nonce))

    def null_queue(self, server_nonce):
        """
        NULL queued_time for remaining entries in queue, to allow
        daemon to take care of them as soon as possible.
        """
        query = """UPDATE queue
                      SET queued = NULL
                    WHERE server_nonce = %s"""
        self._execute(query, (server_nonce,))

    def update_db_counters(self, params):
        """ Update table with new counter values """
        query = """UPDATE yubikeys
                      SET modified = %(modified)s,
                          yk_counter = %(yk_counter)s,
                          yk_use = %(yk_use)s,
                          yk_low = %(yk_low)s,
                          yk_high = %(yk_high)s,
                          nonce = %(nonce)s
                    WHERE yk_publicname = %(yk_publicname)s
                      AND (yk_counter < %(yk_counter)s
                       OR (yk_counter = %(yk_counter)s
                      AND yk_use < %(yk_use)s))"""
        self._execute(query, params)

    def enqueue(self, otp_params, local_params, server, server_nonce):
        """
        Insert new params into database queue table
        """
        info = 'yk_publicname=%(yk_publicname)s&yk_counter=%(yk_counter)s' % otp_params
        info += '&yk_use=%(yk_use)s&yk_high=%(yk_high)s&yk_low=%(yk_low)s' % otp_params
        info += '&nonce=%(nonce)s' % otp_params
        info += ',&local_counter=%(yk_counter)s&local_use=%(yk_use)s' % local_params
        query = """INSERT INTO queue (
                        queued,
                        modified,
                        otp,
                        server,
                        server_nonce,
                        info
                    ) VALUES (%s, %s, %s, %s, %s, %s)"""
        self._execute(query, (int(time.time()), otp_params['modified'],
                              otp_params['otp'], server, server_nonce, info))

    def get_keys(self, yk_publicname):
        """ Get all keys from DB """
        query = """SELECT yk_publicname
                     FROM yubikeys
                    WHERE active = 1"""
        params = None
        if yk_publicname != 'all':
            query += ' AND yk_publicname = %s'
            params = (yk_publicname,)
        self._execute(query, params)
        return self._dictfetchall()

    ################
    # YKKSM QUERIES
    ################
    def get_key_and_internalname(self, public_id):
        """
        Read token's AESkey and internalname for OTP decryption
        """
        query = """SELECT aeskey,
                          internalname
                     FROM yubikeys
                    WHERE (active = '1' OR active = 'true')
                      AND publicname = %s"""
        self._execute(query, (public_id,))
        return self._dictfetchone()



"""
yubistack.config
~~~~~~~~~~~~~~~~

Settings file parser module
"""

__all__ = [
    'settings',
]

import imp
import os
import logging

# CONSTANTS
TS_SEC = 0.128 # 8Hz timer of the HW token
TS_REL_TOLERANCE = 0.3
TOKEN_LEN = 32 # Lengh of Yubikey OTP tokens
OTP_MAX_LEN = 48 # TOKEN_LEN plus public identity of 0..16

SETTINGS_FILE = os.getenv('YUBISTACK_SETTINGS', '/etc/yubistack.conf')
DEFAULT_CRYPT_CONTEXT = {
    'schemes': ['sha256_crypt'],
    'deprecated': [],
    'default': 'sha256_crypt',
    'yhsm_pbkdf2_sha1__key_handle': 1,
    'all__vary_rounds': 0.1,
    'sha256_crypt__min_rounds': 80000,
    'admin__sha256_crypt__min_rounds': 160000
}
VALUES = [
    ('DATABASES', {}),
    ('USE_HSM', False),
    ('USE_NATIVE_YKVAL', False),
    ('USE_NATIVE_YKKSM', False),
    ('CRYPT_CONTEXT', DEFAULT_CRYPT_CONTEXT),
    ('YKVAL_CLIENT_ID', None),
    ('YKVAL_CLIENT_SECRET', ''),
    ('YKVAL_SERVERS', []),
    ('YKKSM_KEYDIR', ''),
    ('YKKSM_SERVERS', []),
    ('YKKSM_KEYDIR', False),
    ('LOGLEVEL', 'INFO'),
    ('LOGFILE', '/tmp/yubistack.log'),
    ('SYNC_SERVERS', []),
    ('SYNC_LEVEL', 100),
    ('SYNC_POOL', []),
    ('SYNC_TIMEOUT', 3),
    ('SYSLOG_WSGI_AUTH', True),
    ('TS_ABS_TOLERANCE', 0),
]

def parse(conf):
    """ Parse settings file parameters into a dict """
    _settings = {}
    for conf_key, default_value in VALUES:
        _settings[conf_key] = getattr(conf, conf_key, default_value)
    return _settings

if os.path.isfile(SETTINGS_FILE):
    user_settings = imp.load_source('user_settings', SETTINGS_FILE)
    settings = parse(user_settings)
else:
    settings = dict(VALUES)

loglevel = getattr(logging, settings['LOGLEVEL'], logging.INFO)
logging.basicConfig(format='%(asctime)s %(levelname)s %(module)s %(funcName)s: %(message)s',
                    level=loglevel, filename=settings['LOGFILE'])



"""
yubistack.ykauth
~~~~~~~~~~~~~~~~

Python Yubikey Stack - Authentication module
"""

import base64
import logging
import urllib

from passlib.context import CryptContext
import requests

from .config import (
    settings,
    TOKEN_LEN,
)
from .db import DBHandler
from .exceptions import YKAuthError
from .utils import (
    sign,
    generate_nonce,
)

logger = logging.getLogger(__name__)


class VerificationClient:
    """ Verification Client """
    def __init__(self, urls, client_id=None, apikey=None):
        self.urls = urls
        self.client_id = client_id
        self.apikey = base64.b64decode(apikey)

    def generate_query(self, otp, nonce, timestamp=False, timeout=None,
                       sync_level=None):
        """ Generate query """
        data = [('id', self.client_id),
                ('otp', otp),
                ('nonce', nonce)]
        if timestamp:
            data.append(('timestamp', '1'))

        if sync_level is not None:
            data.append(('sl', sync_level))

        if timeout:
            data.append(('timeout', timeout))

        query_string = urllib.parse.urlencode(data)
        if self.apikey:
            signature = sign(dict(data), self.apikey)
            query_string += '&h=%s' % (signature.replace('+', '%B'))
        return query_string

    def verify(self, otp, timestamp=False, sl=None, timeout=None,
               return_response=False):
        """ Make a HTTP call to the Yubikey Verification servers """
        nonce = generate_nonce()
        query = self.generate_query(otp, nonce, timestamp=timestamp,
                                    timeout=timeout, sync_level=sl)
        req = requests.get(self.urls[0] + '?' + query)
        print(req.text)


class Client:
    """ Authentication Client """
    def __init__(self):
        self.db = DBHandler(db='yubiauth')
        self.pwd_context = CryptContext(**settings['CRYPT_CONTEXT'])
        if settings['USE_NATIVE_YKVAL']:
            # Native verify
            from .ykval import Validator
            self.ykval_client = Validator()
        else:
            # Using yubico_client to verify against remote server
            from yubico_client import Yubico
            self.ykval_client = Yubico(settings['YKVAL_CLIENT_ID'],
                                       settings['YKVAL_CLIENT_SECRET'],
                                       api_urls=settings['YKVAL_SERVERS'])

    def _get_user_info(self, username):
        """
        Get user from DB

        Args:
            username

        Returns:
            dictionary of user data

        Raises:
            AuthFail if user does not exist
        """
        user = self.db.get_user(username)
        if not user:
            raise YKAuthError('UNKNOWN_USER')
        logger.debug('[%s] Found user: %s', username, user)
        return user

    def _check_token(self, user, token_id):
        """
        Check Token association with user

        Args:
            user: User data dict as recieved from _get_user_info()
            token_id: Token prefix (aka. publicname)

        Returns:
            None

        Raises:
            AuthFail if token is not associated with the user
            AithFail if token is disabled
        """
        token = self.db.get_token(user['users_id'], token_id)
        if not token:
            logger.error('[%s] Token %s is not associated with user',
                         user['users_name'], token_id)
            raise YKAuthError('INVALID_TOKEN')
        logger.debug('[%s] Found token: %s', user['users_name'], token)
        if not token.get('yubikeys_enabled'):
            logger.error('[%s] Token %s is disabled for %s',
                         user['users_name'], token_id, user['users_name'])
            raise YKAuthError('DISABLED_TOKEN')

    def _validate_password(self, user, password):
        """
        Validate password against the hash in SQL
        """
        valid, new_hash = self.pwd_context.verify_and_update(str(password), user['users_auth'])
        if not valid:
            logger.error('[%(users_name)s] Invalid password', user)
            raise YKAuthError('BAD_PASSWORD')
        if new_hash:
            # TODO: update user's hash with new_hash
            logger.warning('[%(users_name)s] User password hash needs update', user)
        return True

    def authenticate(self, username, password, otp):
        """
        Yubistack user authentication

        Args:
            username: Username of the user
            password: Password/PIN of the user
            otp: Yubikey one time password

        Returns:
            dict of authentication data

        Authentication process:
            1. Check if token is enabled
            2. Check if token is associated with the user & enabled
            3. Validate users password
            4. Validate OTP (YKVal)
        """
        token_id = otp[:-TOKEN_LEN]
        # STEP 1: Check if token is enabled
        user = self._get_user_info(username)
        # STEP 2: Check if token is associated with the user & enabled
        self._check_token(user, token_id)
        # STEP 3: Validate users password
        self._validate_password(user, password)
        # STEP 4: Validate OTP
        self.ykval_client.verify(otp)
        return True



"""
ykval.utils
~~~~~~~~~~~

Helper utilities for Yubikey Validator (YK-VAL)
"""

import base64
from binascii import hexlify, unhexlify
from cgi import parse_qs
from Crypto.Cipher import AES
from datetime import datetime
import hashlib
import hmac
import logging
from random import getrandbits
import re
import sys

logger = logging.getLogger(__name__)

HEX_CHARS = '0123456789abcdef'
MODHEX_CHARS = 'cbdefghijklnrtuv'
if sys.version_info < (3,):
    import string
    MODHEX_TO_HEX_MAP = string.maketrans('cbdefghijklnrtuv', '0123456789abcdef')
    HEX_TO_MODHEX_MAP = string.maketrans(HEX_CHARS, MODHEX_CHARS)
else:
    MODHEX_TO_HEX_MAP = str.maketrans('cbdefghijklnrtuv', '0123456789abcdef')
    HEX_TO_MODHEX_MAP = str.maketrans(HEX_CHARS, MODHEX_CHARS)

def parse_querystring(query_string):
    """
    Parse request's query string
    >>> r = parse_querystring('/foo?bar=1&moo=abc&other=-123')
    >>> r == {'/foo?bar': 1, 'moo': 'abc', 'other': -123}
    True
    """
    params = {}
    for key, val in parse_qs(query_string).items():
        if (
                (val[0] == '0') or
                (val[0].isdigit() and not val[0].startswith('0')) or
                (val[0].startswith('-') and val[0][1:].isdigit())
            ):
            params[key] = int(val[0])
        else:
            params[key] = val[0]
    return params

def modhex2hex(modhex):
    """
    Turn modhex into regular hex
    >>> modhex2hex('aabbccddeeffnrfrjeelfvehntlnlvgejvcrtdvivfbh')
    'aa1100223344bc4c833a4f36bdabaf538f0cd2f7f416'
    """
    # Python2 fails to translate unicode
    return str(modhex).translate(MODHEX_TO_HEX_MAP)

def hex2modhex(_hex):
    """
    Turn regular hex into modhex
    >>> hex2modhex('aa1100223344bc4c833a4f36bdabaf538f0cd2f7f416')
    'llbbccddeeffnrfrjeelfvehntlnlvgejvcrtdvivfbh'
    """
    # Python2 fails to translate unicode
    return str(_hex).translate(HEX_TO_MODHEX_MAP)

def aes128ecb_decrypt(key, cipher):
    """
    Decrypt ciphertext with key using AES128CBC encryption
    >>> key = 'abcdef0123456789abcdef0123456789'
    >>> cipher = '16e1e5d9d3991004452007e302000000'
    >>> plain = aes128ecb_decrypt(key, cipher).decode()
    >>> plain == '46b029d5340bbd23b39c6c9154d095b1'
    True
    """
    key = unhexlify(key)
    iv = unhexlify('00000000000000000000000000000000')
    decryptor = AES.new(key, AES.MODE_ECB, iv)
    cipher = modhex2hex(cipher)
    cipher = unhexlify(cipher)
    plain = decryptor.decrypt(cipher)
    plain = hexlify(plain)
    return plain

def aes128ecb_encrypt(key, plain):
    """
    Decrypt ciphertext with key using AES128CBC encryption
    >>> key = 'abcdef0123456789abcdef0123456789'
    >>> plain = '46b029d5340bbd23b39c6c9154d095b1'
    >>> cipher = aes128ecb_encrypt(key, plain)
    >>> cipher == '16e1e5d9d3991004452007e302000000'
    True
    """
    key = unhexlify(key)
    iv = unhexlify('00000000000000000000000000000000')
    encryptor = AES.new(key, AES.MODE_ECB, iv)
    plain = plain.encode() if not isinstance(plain, bytes) else plain
    plain = unhexlify(plain)
    cipher = encryptor.encrypt(plain)
    cipher = hexlify(cipher)
    cipher = hex2modhex(cipher.decode())
    return cipher

def calculate_crc(token):
    """
    Calculate CRC-16
    >>> calculate_crc('16e1e5d9d3991004452007e302000000')
    22744
    """
    crc = 0xffff
    token = unhexlify(token)
    for i in range(len(token)):
        if isinstance(token[i], str):
            # Python 2.X
            crc = crc ^ ord(token[i])
        else:
            crc = crc ^ token[i]
        for _ in range(8):
            last_bit_true = crc & 0x1
            crc = crc >> 1
            if last_bit_true:
                crc = crc ^ 0x8408
    return crc

def check_crc(token):
    """ Check the value of the CRC function """
    return calculate_crc(token) == 0xf0b8

def sign(data, apikey):
    """
    Sign a http query string in the array of key-value pairs
    return b64 encoded hmac hash

    https://github.com/Yubico/yubikey-val/blob/master/doc/Validation_Protocol_V2.0.adoc#generating-signatures
    """
    # Alphabetically sort the set of key/value pairs by key order.
    keys = sorted(data.keys())
    # Construct a single line with each ordered key/value pair concatenated using &,
    # and each key and value contatenated with =. Do not add any linebreaks.
    # Do not add whitespace. For example: a=2&b=1&c=3.
    query_string = '&'.join(['%s=%s' % (k, data[k]) for k in keys])
    # Apply the HMAC-SHA-1 algorithm on the line as an octet string using the API key as key
    signature = hmac.new(apikey, query_string.encode(), hashlib.sha1).digest()
    # Base 64 encode the resulting value according to RFC 4648
    signature = base64.b64encode(signature).decode()
    logger.debug('Signed data: %s (H=%s)', query_string, signature)
    return signature

def generate_nonce():
    """
    Generate a random nonce
    """
    return hex(getrandbits(128))[2:-1]

def counters_eq(params1, params2):
    """
    Function to compare two set of parameters
    >>> p1 = {'yk_counter': 123, 'yk_use': 200}
    >>> p2 = {'yk_counter': 123, 'yk_use': 200}
    >>> counters_eq(p1, p2)
    True
    >>> p2 = {'yk_counter': 1, 'yk_use': 200}
    >>> counters_eq(p1, p2)
    False
    """
    return params1['yk_counter'] == params2['yk_counter'] \
        and params1['yk_use'] == params2['yk_use']

def counters_gt(params1, params2):
    """
    Function to compare two set of parameters
    >>> p1 = {'yk_counter': 123, 'yk_use': 200}
    >>> p2 = {'yk_counter': 123, 'yk_use': 200}
    >>> counters_gt(p1, p2)
    False
    >>> p2 = {'yk_counter': 80, 'yk_use': 200}
    >>> counters_gt(p1, p2)
    True
    >>> p2 = {'yk_counter': 80, 'yk_use': 201}
    >>> counters_gt(p1, p2)
    True
    """
    return params1['yk_counter'] > params2['yk_counter'] \
        or (params1['yk_counter'] == params2['yk_counter'] \
        and params1['yk_use'] > params2['yk_use'])

def counters_gte(params1, params2):
    """
    Function to compare two set of parameters
    >>> p1 = {'yk_counter': 123, 'yk_use': 200}
    >>> p2 = {'yk_counter': 123, 'yk_use': 200}
    >>> counters_gte(p1, p2)
    True
    >>> p2 = {'yk_counter': 80, 'yk_use': 200}
    >>> counters_gte(p1, p2)
    True
    >>> p1 = {'yk_counter': 80, 'yk_use': 201}
    >>> counters_gte(p1, p2)
    True
    """
    return params1['yk_counter'] > params2['yk_counter'] \
        or (params1['yk_counter'] == params2['yk_counter'] \
        and params1['yk_use'] >= params2['yk_use'])

SYNC_RESPONSE_CHECKS = {
    'modified': re.compile(r'(-1|\d+)'),
    'yk_publicname': re.compile(r'[cbdefghijklnrtuv]+'),
    'yk_counter': re.compile(r'(-1|\d+)'),
    'yk_use': re.compile(r'(-1|[0-9]+)'),
    'yk_high': re.compile(r'(-1|[0-9]+)'),
    'yk_low': re.compile(r'(-1|[0-9]+)'),
    'nonce': re.compile(r'[a-zA-Z0-9]+'),
}
def parse_sync_response(sync_response):
    """ Parsing query string parameters into a dict """
    params = [line.split('=', 1) for line in sync_response.split('\r\n') if '=' in line]
    params = dict(params)
    for name, regex in SYNC_RESPONSE_CHECKS.items():
        if name not in params or not re.match(regex, params[name]):
            raise ValueError('Cannot parse "%s". Response from sync server:\n%s' %
                             (name, sync_response))
        else:
            if params[name].isdigit() or params[name] == '-1':
                params[name] = int(params[name])
    return params

def wsgi_response(resp, start_response, apikey=b'', extra=None, status=200):
    """ Function to return a proper WSGI response """
    # yubikey-val's getUTCTimeStamp() function...
    now = datetime.utcnow().isoformat().replace('.', 'Z')[:-2]
    resp_data = {'status': resp, 't': now}
    if not extra:
        extra = {}
    for key, val in extra.items():
        resp_data[key] = val
    signature = sign(resp_data, apikey)

    # Bulding response string
    response = 'h=%s\r\n' % signature
    response += 't=%s\r\n' % now
    for key, val in extra.items():
        response += '%s=%s\r\n' % (key, val)
    response += 'status=%s\r\n' % resp
    response += '\r\n'
    logger.debug('Response: %s', response)
    start_response('%d OK' % status, [('Content-Type', 'text/plain')])
    return [response.encode()]

if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
yubistack.ykksm
~~~~~~~~~~~~~~~

Yubikey Key Storage Manager - Decryption module
"""

import base64
import logging
import re

from Crypto.Random import atfork

from .config import settings
from .crypt import PrivateKey
from .db import DBHandler
from .exceptions import YKKSMError
from .utils import (
    aes128ecb_decrypt,
    check_crc,
)

logger = logging.getLogger(__name__)
OTP_REGEX = re.compile(r'^(?P<yk_id>[cbdefghijklnrtuv]{0,16})(?P<modhex>[cbdefghijklnrtuv]{32})$')


CRYPTER = None
if settings.get('YKKSM_KEYDIR') and PrivateKey:
    CRYPTER = PrivateKey(settings['YKKSM_KEYDIR'])

class DecryptorDBH(DBHandler):
    """
    Transparently decrypt RSA encrypted aeskeys
    """
    def get_key_and_internalname(self, public_id):
        data = DBHandler.get_key_and_internalname(self, public_id)
        if data:
            try:
                # AES keys are 16 bytes (hex)
                if len(data['aeskey']) != 32:
                    raise ValueError
                else:
                    int(data['aeskey'], 16)
            except ValueError:
                # Re-initialize RNG to deal with forks in uwsgi
                atfork()
                # Encrypted ciphertext with base64 encoding
                ciphertext = base64.b64decode(data['aeskey'])
                data['aeskey'] = CRYPTER.decrypt(ciphertext)

            return data


class Decryptor:
    """ Object to decrypt an OTP """
    def __init__(self, db='ykksm'):
        if settings.get('YKKSM_KEYDIR'):
            self.db = DecryptorDBH(db=db)
        else:
            self.db = DBHandler(db=db)

    def _parse_otp(self, otp):
        """
        Parse Yubikey OTP and return (public_id, modhex)

        Example:
        >>> d._parse_otp('ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded')
        ('ccccccbcgujh', 'ingjrdejhgfnuetrgigvejhhgbkugded')
        """
        match = re.match(OTP_REGEX, otp)
        if not match:
            logger.error('Invalid OTP format: %s', otp)
            raise YKKSMError('BAD_OTP')
        return match.groups()

    def _get_key_and_internalname(self, public_id):
        """
        Get user's AES key from DB

        Output: (aeskey, internalname)

        NOTE: Consider using a Yubikey HSM stick instead of ykksm
        """
        data = self.db.get_key_and_internalname(public_id)
        if not data:
            raise YKKSMError('UNKNOWN_TOKEN')
        logger.debug('Found user: ID: %s, INTERNALNAME: %s',
                     public_id, data['internalname'])
        return (data['aeskey'], data['internalname'])

    def decrypt(self, otp):
        """ Decrypt OTP """
        if not otp:
            logger.error('Missing OTP')
            raise YKKSMError('MISSING_OTP')
        # Get public_id, modhex, key and internalname
        public_id, modhex = self._parse_otp(otp)
        aeskey, internalname = self._get_key_and_internalname(public_id)
        # Get plaintext from key + modhexed cipher
        plaintext = aes128ecb_decrypt(aeskey, modhex)
        plaintext = plaintext.decode()
        # Check for plaintext corruption
        if plaintext[:12] != internalname:
            logger.error('UID Error: %s %s: %s vs %s',
                         otp, plaintext, plaintext[:12], internalname)
            raise YKKSMError('CORRUPT_OTP')
        if not check_crc(plaintext):
            logger.error('CRC Error: %s: %s', otp, plaintext)
            raise YKKSMError('CORRUPT_OTP')
        # Construct output
        output = {
            'counter': plaintext[14:16] + plaintext[12:14],
            'low': plaintext[18:20] + plaintext[16:18],
            'high': plaintext[20:22],
            'use': plaintext[22:24]
        }
        logger.debug('SUCCESS OTP %s PT %s %s', otp, plaintext, output)
        return output



"""

Yubikey Python Authentication Stack
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Components:
    ykauth: Authentication Module
    ykval: Validation Module
    ykksm: Key Storage Module

Examples:
>>> from yubistack import Client
>>> c = Client()
>>> c.authenticate('joe', 'secret', 'ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded')
False
>>>
>>> from yubistack import Validator
>>> v = Validator()
>>> v.verify(1, 'ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded')
{'otp': 'ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded', 'time': '2015-04-14T20:07:20Z5261', \
 'nonce': None, 'status': 'OK', 'sl': None, 'signature': 'h0P2wfJUqHQFuRpG4n1Kvk3KacE='}
>>>
>>> from yubistack import Decryptor
>>> d = Decryptor()
>>> d.decrypt('ccccccbcgujhingjrdejhgfnuetrgigvejhhgbkugded')
{'counter': '0235', 'low':' 21c4', 'high': 'b2', 'use':'42'}
"""

__version__ = '0.5.4'
__all__ = [
    'settings',
    'Client',
    'Decryptor',
    'Sync',
    'Validator',
]

import imp
import os
import logging

from .config import settings
from .ykauth import Client
from .ykksm import Decryptor
from .yksync import Sync
from .ykval import Validator



"""
yubistack.ykval
~~~~~~~~~~~~~~~

Python Yubikey Stack - Key validation module
"""

import base64
from datetime import datetime
import logging
import re
import time

import requests

from .config import (
    settings,
    TS_SEC,
    TS_REL_TOLERANCE,
    TOKEN_LEN,
    OTP_MAX_LEN,
)
from .db import DBHandler
from .exceptions import YKValError
from .utils import (
    generate_nonce,
    counters_eq,
    counters_gte,
)
from .yksync import Sync

logger = logging.getLogger(__name__)

class Validator:
    """ Yubikey OTP validator """
    def __init__(self):
        self.db = DBHandler(db='ykval')
        if settings['USE_NATIVE_YKKSM']:
            from .ykksm import Decryptor
            self.decryptor = Decryptor()
        else:
            self.decryptor = None
        self.sync_servers = settings['SYNC_SERVERS']
        self.default_sync_level = settings['SYNC_LEVEL']
        # Below parameters are valid from protocol version >= 2.0
        self.sync_level = None
        self.timeout = None

    def check_parameters(self, params):
        """ Perform Sanity check on parameters """
        params['token_id'] = '?' * 12
        # OTP
        if not TOKEN_LEN <= len(params['otp']) <= OTP_MAX_LEN:
            logger.error('[%(token_id)s] Incorrect OTP length: %(otp)s', params)
            raise YKValError('BAD_OTP')
        params['token_id'] = params['otp'][:-TOKEN_LEN]
        if not re.match(r'^[cbdefghijklnrtuv]+$', params['otp']):
            logger.error('[%(token_id)s] Invalid OTP: %(otp)s', params)
            raise YKValError('BAD_OTP')
        # CLIENT ID
        if params['client_id'] and not isinstance(params['client_id'], int):
            logger.error('[%(token_id)s] id provided in request '
                         '(%(client_id)s must be an integer', params)
            raise YKValError('INVALID_PARAMETER', 'client_id')
        # NONCE:
        # - If client_id is not provided, we're using a Native stack call
        if params['client_id'] and not params['nonce']:
            logger.error('[%(token_id)s] Nonce is missing', params)
            raise YKValError('MISSING_PARAMETER', 'nonce')
        if params['nonce'] and not re.match(r'^[A-Za-z0-9]+$', params['nonce']):
            logger.error('[%(token_id)s] Nonce is provided but not correct', params)
            raise YKValError('INVALID_PARAMETER', 'nonce')
        if params['nonce'] and not 16 <= len(params['nonce']) <= 40:
            logger.error('[%(token_id)s] Nonce too short or too long (%(nonce)s)', params)
            raise YKValError('INVALID_PARAMETER', 'nonce')
        # TIMESTAMP
        #   NOTE: Timestamp parameter is not checked since current protocol says
        #   that 1 means request timestamp and anything else is discarded.
        # TIMEOUT
        if not isinstance(params['timeout'], int):
            logger.error('[%(token_id)s] timeout is provided but not correct (%(timeout)s)',
                         params)
            raise YKValError('INVALID_PARAMETER', 'timeout')
        # SYNC LEVEL
        if not (isinstance(params['sync_level'], int) and 0 <= params['sync_level'] <= 100):
            logger.error('[%(token_id)s] SL (sync level) is provided but '
                         'not correct (%(sync_level)s)', params)
            raise YKValError('INVALID_PARAMETER', 'sync_level')

    def get_client_apikey(self, client_id):
        """
        Get Client info from DB

        Args:
            client_id: Integer ID number of the client.
                       Corresponding API key will be retrieved from DB

        Returns:
            b64decoded client secret (apikey)

        Raises:
            YKValError('NO_SUCH_CLIENT') if client doesn't exist
        """
        if not client_id:
            return ''.encode()
        client_data = self.db.get_client_data(client_id)
        logger.debug('Client data: %s', client_data)
        if not client_data:
            logger.error('Invalid client id: %s', client_id)
            raise YKValError('NO_SUCH_CLIENT')
        return base64.b64decode(client_data['secret'])

    def decode_otp(self, otp):
        """
        Call out to KSM to decrypt OTP
        """
        if self.decryptor:
            data = self.decryptor.decrypt(otp)
            return dict([(k, int(v, 16)) for k, v in data.items()])
        elif settings['YKKSM_SERVERS']:
            # TODO: Support for async req for multiple servers
            for url in settings['YKKSM_SERVERS']:
                req = requests.get(url, params={'otp': otp}, headers={'Accept': 'application/json'})
                logger.debug('[%s] YK-KSM response: %s (status_code: %s)',
                             otp[:-TOKEN_LEN], req.text, req.status_code)
                if req.headers['Content-Type'] == 'application/json' and req.status_code == 200:
                    return dict([(k, int(v, 16)) for k, v in req.json().items()])
                if req.text.startswith('OK'):
                    resp = {}
                    for i in req.text.split()[1:]:
                        key, val = i.split('=')
                        resp[key] = int(val, 16)
                    return resp
            raise YKValError('BAD_OTP')
        logger.error("No KSM service provided. Can't decrypt OTP.")
        raise YKValError('BACKEND_ERROR', 'No KSM service found')

    def build_otp_params(self, params, otp_info):
        """ Build OTP params """
        return {
            'modified': int(time.time()),
            'otp': params['otp'],
            'nonce': params['nonce'],
            'yk_publicname': params['otp'][:-TOKEN_LEN],
            'yk_counter': int(otp_info['counter']),
            'yk_use': int(otp_info['use']),
            'yk_high': otp_info['high'],
            'yk_low': otp_info['low'],
        }

    def validate_otp(self, otp_params, local_params):
        """ Validate OTP """
        # First check if OTP is seen with the same nonce, in such case we have an replayed request
        if counters_eq(local_params, otp_params) and local_params['nonce'] == otp_params['nonce']:
            logger.error('[%(yk_publicname)s] Replayed request '
                         '(OTP: %(otp)s, Nonce: %(nonce)s)', otp_params)
            raise YKValError('REPLAYED_REQUEST')
        # Check the OTP counters against local DB
        if counters_gte(local_params, otp_params):
            logger.error('[%s] Replayed OTP: Local counters higher (%s > %s)',
                         otp_params['yk_publicname'], local_params, otp_params)
            raise YKValError('REPLAYED_OTP')
        # Valid OTP, update DB
        self.db.update_db_counters(otp_params)

    def replicate(self, otp_params, local_params, server_nonce):
        """ Handle sync across the cluster """
        for server in settings['SYNC_SERVERS']:
            self.db.enqueue(otp_params, local_params, server, server_nonce)

        req_answers = round(len(self.sync_servers) * float(self.sync_level) / 100.0)
        if req_answers:
            sync = Sync(self.db)
            sync_metrics = sync.sync_remote(otp_params, local_params, server_nonce,
                                            req_answers, self.timeout)
            sync_success = sync_metrics['valid_answers'] == req_answers
            sync_level_success_rate = 100.0 * sync_metrics['valid_answers'] / len(self.sync_servers)
        else:
            sync_success = True
            sync_level_success_rate = 0
            sync_metrics = {'answers': 0, 'valid_answers': 0}
        logger.info('[%s] Sync details: synclevel=%s server_count=%s req_answers=%s '
                    'answers=%s valid_answers=%s sl_success_rate=%.3f timeout=%s',
                    otp_params['yk_publicname'], self.sync_level, len(self.sync_servers),
                    req_answers, sync_metrics['answers'], sync_metrics['valid_answers'],
                    sync_level_success_rate, self.timeout)

        if not sync_success:
            # sync returned false, indicating that either at least 1 answer
            # marked OTP as invalid or there were not enough answers
            if sync_metrics['valid_answers'] != sync_metrics['answers']:
                logger.error('[%(yk_publicname)s] Remote server claims Replayed '
                             'request (OTP: %(otp)s, Nonce: %(nonce)s)', otp_params)
                raise YKValError('REPLAYED_OTP')
            else:
                logger.error('[%s] Failed to synchronize with %s '
                             'servers (%s answers, %s valid)', otp_params['yk_publicname'],
                             req_answers, sync_metrics['answers'], sync_metrics['valid_answers'])
                raise YKValError('NOT_ENOUGH_ANSWERS')
        return sync_level_success_rate

    def phishing_test(self, otp_params, local_params):
        """
        Run a test against token's internal timer

        If the token was generated seconds before the login attempt
        mark the OTP as invalid.
        """
        # Only check token timestamps if TS_ABS_TOLERANCE is
        # set to a proper value or token was not plugged out
        if (
                settings['TS_ABS_TOLERANCE'] == 0 or
                otp_params['yk_counter'] != local_params['yk_counter']
            ):
            return
        new_ts = (otp_params['yk_high'] << 16) + otp_params['yk_low']
        old_ts = (local_params['yk_high'] << 16) + local_params['yk_low']
        ts_delta = (new_ts - old_ts) * TS_SEC

        # Check real time
        last_time = local_params['modified']
        now = int(time.time())
        elapsed = now - last_time
        deviation = abs(elapsed - ts_delta)

        # Time delta server might verify multiple OTPs in a row. In such case validation server
        # doesn't have time to tick a whole second and we need to avoid division by zero.
        if elapsed:
            percent = deviation / elapsed
        else:
            percent = 1
        if deviation > settings['TS_ABS_TOLERANCE'] and percent > TS_REL_TOLERANCE:
            logger.error('[%s] OTP Expired: TOKEN TS DIFF: %s, '
                         'ACCESS TS DIFF: %s, DEVIATION: %s (sec) or %s%%',
                         otp_params['yk_publicname'], ts_delta, elapsed, deviation, percent)
            raise YKValError('DELAYED_OTP')

    def verify(self, otp, client_id=None, nonce=None, timestamp=0,
               timeout=None, sync_level=None):
        """
        Yubico OTP Validation Protocol V2.0 Implementation

        Args:
            otp: The OTP from the YubiKey
            client_id: Specifies the requestor so that the end-point
                       can retrieve correct shared secret for
                       signing the response.
            nonce: A 16 to 40 character long string with random unique data
            timestamp: Timestamp=1 requests timestamp and session
                       counter information in the response
            timeout: Number of seconds to wait for sync responses;
                     if absent, let the server decide
            sync_level: A value 0 to 100 indicating percentage of
                        syncing required by client, or strings "fast" or
                        "secure" to use server-configured values;
                        if absent, let the server decide

        Returns:
            A signed response with status=OK if the OTP is valid

        Raises:
            YKValError('BAD_OTP'): The OTP is invalid format
            YKValError('REPLAYED_OTP'): The OTP has already been seen by the service
            YKValError('MISSING_PARAMETER'): The request lacks a parameter
            YKValError('NO_SUCH_CLIENT'): The request id does not exist
            YKValError('BACKEND_ERROR'): Unexpected error in the server
            YKValError('NOT_ENOUGH_ANSWERS'): Server could not get requested number
                                              of syncs during before timeout
            YKValError('REPLAYED_REQUEST'): Server has seen the OTP/Nonce combination before


        Verify OTP process:
            1. sanitize input parameters
            2. decrypt OTP (YKKSM)
            3. compare old OTP counters with the given OTP counters and check for replay
            4. replicate new OTP counters to remote servers and check for replay on other servers
            5. check for phishing: OTP has to be used within a timeframe otherwise mark as expired
            6. prepare response: Sign the response with the right client key
        """

        ###################################
        # STEP 1: sanitize input parameters
        ###################################
        self.timeout = timeout if timeout else settings['SYNC_TIMEOUT']
        self.sync_level = sync_level if sync_level else settings['SYNC_LEVEL']
        server_nonce = generate_nonce()
        params = {
            'client_id': client_id,
            'otp': otp,
            'nonce': nonce if nonce else server_nonce,
            'timestamp': timestamp,
            'timeout': self.timeout,
            'sync_level': self.sync_level,
        }
        extra_params = {
            'otp': otp,
            'nonce': params['nonce']
        }
        # Check sanity of parameters
        self.check_parameters(params)

        #####################
        # STEP 2: decrypt OTP
        #####################
        otp_info = self.decode_otp(otp)

        #######################################
        # STEP 3: compare old OTP counters with
        #         the given OTP counters and
        #         check for replay
        #######################################
        # Get old parameters (counters) for the token
        local_params = self.db.get_local_params(otp[:-TOKEN_LEN])
        if not local_params['active']:
            logger.error('[%(yk_publicname)s]: De-activated Yubikey', local_params)
            raise YKValError('DISABLED_TOKEN')
        # Build the new parameters (counters) for the given OTP
        otp_params = self.build_otp_params(params, otp_info)
        # Validate OTP, check for replayed request or replayed OTP
        self.validate_otp(otp_params, local_params)

        #####################################
        # STEP 4: replicate new OTP counters
        #         to remote servers and check
        #         for replay on other servers
        #####################################
        sync_level_success_rate = self.replicate(otp_params, local_params, server_nonce)

        #######################################
        # STEP 5: check for phishing, OTP has
        #         to be used within a timeframe
        #         otherwise mark as expired
        #######################################
        self.phishing_test(otp_params, local_params)

        ##########################
        # STEP 6: Prepare response
        ##########################
        extra_params['sl'] = sync_level_success_rate
        if timestamp == 1:
            extra_params['timestamp'] = (otp_info['yk_high'] << 16) + otp_info['yk_low']
            extra_params['sessioncounter'] = otp_info['yk_counter']
            extra_params['sessionuse'] = otp_info['yk_use']
        response = {
            'status': 'OK',
            'time': datetime.utcnow().isoformat().replace('.', 'Z')[:-2]
        }
        response.update(extra_params)
        return response



"""
yubistack.softtoken
~~~~~~~~~~~~~~~~~~~

Generating OTP with code o/
"""

from random import randint

from .utils import (
    calculate_crc,
    aes128ecb_encrypt,
)

class SoftToken:
    """
    SoftToken generator class

    Usage:
    >>> st = SoftToken('tetetetetecc', '8792ebfe26cc', 'ecde18dbe76fbd0c33330f1c354871db')
    >>> st.counter = 19
    >>> st.timestamp_low = 49320
    >>> st.timestamp_high = 0
    >>> st.session_use = 16
    >>> for i in range(0, 5):
    ...     st.otp()
    ...
    'tetetetetecccrgkcdjhtfebjvvvufuuccvetbinrhuv'
    'teteteteteccvtjfbfdrictrndbeubdtccbhhdblinvb'
    'teteteteteccfjtjkfhrnbcrgjnrcvrktffnirdkhukn'
    'teteteteteccfjjkiivgrcnvbkgiffckkglentvtekfb'
    'teteteteteccvhfftjitevuvuehhbvcktcgndnuiblbr'
    >>>
    """

    def __init__(self, public_id, internalname, aeskey):
        self.public_id = public_id
        self.internalname = internalname
        self.aeskey = aeskey
        self.counter = 0
        self.timestamp_low = 0
        self.timestamp_high = 0
        self.session_use = 0

    def otp(self):
        """ Generate OTP token """
        _random = '%.4x' % randint(0, 65536)
        counter_hex = '%.4x' % self.counter
        low_hex = '%.4x' % self.timestamp_low
        high_hex = '%.2x' % self.timestamp_high
        use_hex = '%.2x' % self.session_use
        self.session_use += 1
        token = (self.internalname +
                 counter_hex[2:4] +
                 counter_hex[0:2] +
                 low_hex[2:4] +
                 low_hex[0:2] +
                 high_hex +
                 use_hex +
                 _random[2:4] +
                 _random[0:2])
        crc_hex = '%.4x' % (~calculate_crc(token) & 0xffff)
        token += crc_hex[2:4] + crc_hex[0:2]
        return self.public_id + aes128ecb_encrypt(self.aeskey, token)

def main():
    """ Main program """
    token = SoftToken('tetetetetecc', '8792ebfe26cc', 'ecde18dbe76fbd0c33330f1c354871db')
    token.counter = 19
    token.timestamp_low = 49320
    token.timestamp_high = 0
    token.session_use = 16
    for _ in range(0, 5):
        print(token.otp())

if __name__ == '__main__':
    main()




"""

yubistack.exceptions
~~~~~~~~~~~~~~~~~~~~

List all custom exceptions here
"""

STATUS_CODES = {
    # YKAuth
    'BAD_PASSWORD': 'Invalid password',
    'DISABLED_TOKEN': 'Token is disabled',
    'UNKNOWN_USER': 'Unknown user',
    'INVALID_TOKEN': 'Token is not associated with user',
    # YKVal
    'BACKEND_ERROR': 'Backend error',
    'BAD_OTP': 'Invalid OTP',
    'BAD_SIGNATURE': 'The HMAC signature verification failed',
    'DELAYED_OTP': 'Expired OTP',
    'INVALID_PARAMETER': 'The request has invalid parameter',
    'MISSING_PARAMETER': 'The request missing parameter',
    'NO_SUCH_CLIENT': 'The request id does not exist',
    'NOT_ENOUGH_ANSWERS': 'Server could not get requested number of syncs before timeout',
    'OPERATION_NOT_ALLOWED': 'The request is now allowed',
    'REPLAYED_OTP': 'Replayed OTP',
    'REPLAYED_REQUEST': 'Server has seen the OTP/Nonce combination before',
    # YKKSM
    'CORRUPT_OTP': 'Corrupt OTP',
    'MISSING_OTP': 'No OTP provided',
    'UNKNOWN_TOKEN': 'Unknown yubikey',
}

class YubistackError(Exception):
    """ Yubistack Exception """
    NAME = 'Yubistack error'
    def __init__(self, *args):
        super(YubistackError, self).__init__(*args)
        self.error_code = self.args[0]

    def __str__(self):
        message = STATUS_CODES[self.error_code]
        if len(self.args) == 2:
            message += ': %s' % self.args[1]
        return message

class YKAuthError(YubistackError):
    """ Error returned by the Client class """
    NAME = 'Authentication error'

class YKValError(YubistackError):
    """ Error returned by the Validator class """
    NAME = 'Validation error'

class YKSyncError(YubistackError):
    """ Error returned by the Sync class """
    NAME = 'Sync error'

class YKKSMError(YubistackError):
    """ Error returned by the Decryptor class """
    NAME = 'Decryption error'



"""
yubistack.replicate
~~~~~~~~~~~~~~~~~~~

Script to run through the queue and ensure
that remote servers are updated.
"""

import logging

import requests

from yubistack.utils import (
    counters_eq,
    counters_gt,
    parse_querystring,
    parse_sync_response,
)
from yubistack.ykval import Sync

logger = logging.getLogger(__name__)
requests_log = logging.getLogger('requests')
requests_log.setLevel(logging.WARNING if logger.getEffectiveLevel() != 10 else logging.DEBUG)

def main():
    """
    Main program
    """
    sync = Sync()
    queue_data = sync.db.read_queue()
    servers = set([x['server'] for x in queue_data])

    for server in servers:
        logger.debug('Processing queue for server %s', server)
        items = [x for x in queue_data if x['server'] == server]
        try:
            for item in items[:20]:
                url = '%s?otp=%s&modified=%s&%s' % (item['server'], item['otp'], item['modified'],
                                                    item['info'].split(',')[0])
                resp = requests.get(url)
                if resp.status_code == 200:
                    if resp.text.rstrip().endswith('status=OK'):
                        # Read remote counters
                        remote_params = parse_sync_response(resp.text)
                        # Update local counters
                        sync.db.update_db_counters(remote_params)
                        # Get OTP counters
                        otp_params = parse_querystring(item['info'].split(',')[0])
                        # Get validation counters before processing the OTP
                        validation_params = parse_querystring(item['info'].split(',')[1])
                        validation_params = {'yk_counter': validation_params['local_counter'],
                                             'yk_use': validation_params['local_use']}
                        # Get current local counters
                        local_params = sync.db.get_local_params(otp_params['yk_publicname'])

                        if counters_gt(validation_params, remote_params):
                            logger.info('[%s]: Remote server out of sync compared to counters '
                                        'at validation request time.', server)
                        if counters_gt(remote_params, validation_params):
                            if counters_eq(remote_params, otp_params):
                                logger.info('[%s]: Remote server had received the current '
                                            'counter values already.', server)
                            else:
                                logger.info('Local server out of sync compared to counters '
                                            'at validation request time.')
                        if counters_gt(local_params, remote_params):
                            logger.info('Local server out of sync compared to current local '
                                        'counters. Local server updated.')
                        if counters_gt(remote_params, otp_params):
                            logger.info('[%s]: Remote server has higher counters than OTP. '
                                        'This response would have marked the OTP as invalid.',
                                        server)
                        elif counters_eq(remote_params, otp_params) and \
                                remote_params['nonce'] != otp_params['nonce']:
                            logger.info('[%s]: Remote server has equal counters as OTP '
                                        'and nonce differs. This response would have '
                                        'marked the OTP as invalid.', server)
                        # Delete queue entry
                        sync.db.remove_from_queue(server, item['modified'], item['server_nonce'])
                    elif resp.text.rstrip().endswith('status=BAD_OTP'):
                        logger.warning('[%s]: Remote server says BAD_OTP, pointless to try '
                                       'again, removing from queue.', server)
                        # Delete queue entry
                        sync.db.remove_from_queue(server, item['modified'], item['server_nonce'])
                    else:
                        logger.error('[%s]: Remote server refused our sync request. '
                                     'Check remote server logs.', server)
                else:
                    logger.error('[%s]: Remote server refused our sync request. '
                                 'Check remote server logs.', server)
        except (requests.exceptions.Timeout,
                requests.exceptions.ConnectionError) as err:
            logger.error('Failed to connect to server %s: %s', server, err)
            continue

if __name__ == '__main__':
    main()
